import vnmrjpy as vj
import numpy as np
from nipype.interfaces import fsl
import os
import time
import nibabel as nib
import copy
import shutil
import warnings

def COMPOSER(varr, varr_ref, workdir=None, keepfiles=False, log=True):
    """Phase match multiple receiver channels with help from a reference scan

    Args:
        varr (vj.varray) -- input sequence to phase match 
        varr_ref (vj.varray) -- short echo time reference scan
        workdir (boolean)  -- /path/to/temp/workdir
                                (if None, defaults to one in vj.config)
        keepfiles (boolean) -- set True to keep work files saved in [workdir]
        log (boolean) -- write log file 
    Return:
        varr (vj.varray) -- same as input varr, but with phase matched data

    Outline:
        Heavy use of FSL FLIRT and applyxfm via nipype. See ref for method

    Ref:
        Robinson et al.: Combining Phase Images from Array Coils Using a
        Short Echo Time Reference Scan (COMPOSER), (2017)
    """
    #sanity check
    if varr.vdtype == 'imagespace' and varr_ref.vdtype == 'imagespace':
        pass
    else:
        raise(Exception('Input data should be in image space'))
    # check reference echo time: is it short?
    et = float(varr_ref.pd['te'])*1000
    if  et > 2:
        print('Ref echo time is '+str(et)+' ms. Might bee too long.')

    rcvr_dim = int(vj.config['rcvr_dim'])
    time_dim = int(vj.config['et_dim'])

    if workdir == None:
        timestr = time.strftime("%Y%m%d-%H%M%S")
        workdir = vj.config['fsl_workdir']+'/composer_'+timestr
    # ------------------------FLIRT registration-------------------------------
    
    #ssos, save to nifti, register varr_ref to varr
    img_out = workdir+'/composer_img'
    ref_out = workdir+'/composer_ref'
    flirt_out = workdir+'/composer_flirtout'
    fullref_out = workdir+'/composer_fullimg'
    matrix_file = workdir+'/composer_invol2refvolmat'
    finref_out = workdir+'/composer_finrefout'
    composer_out = workdir+'/composer_finalout'
    # flirt cannot handle more than 3D, so cut data
    
    vj.core.write_nifti(varr,img_out,cut_to_3d=True)
    vj.core.write_nifti(varr_ref,ref_out, cut_to_3d=True)
    vj.core.write_nifti(varr_ref,fullref_out,\
                save_complex=True,rcvr_to_timedim=True,\
                combine_rcvrs=False)
    #register short echo time reference to main image
    flirt = fsl.FLIRT()
    flirt.inputs.in_file = ref_out+'.nii.gz'
    flirt.inputs.reference = img_out+'.nii.gz'
    flirt.inputs.out_file = flirt_out+'.nii.gz'
    flirt.inputs.out_matrix_file = matrix_file
    flirt.inputs.interp = 'spline'
    res = flirt.run()
    # apply transform matrix to reference, but keep each channel
    applyxfm = fsl.preprocess.ApplyXFM()
    applyxfm.inputs.in_file = fullref_out+'.nii.gz'
    applyxfm.inputs.in_matrix_file = matrix_file
    applyxfm.inputs.out_file = finref_out+'.nii.gz'
    applyxfm.inputs.out_matrix_file = workdir+'/applyxfm_matrix'
    applyxfm.inputs.reference = img_out+'.nii.gz'
    applyxfm.inputs.interp = 'nearestneighbour'
    applyxfm.inputs.apply_xfm=True
    res = applyxfm.run()
    # load nifti data with nibabel
    data = nib.load(finref_out+'.nii.gz').get_fdata()
    # extract phase from aligned short echo time reference:
    ph_setr = data[:,:,:,4:]
    # align dimensions, so rcvr is dim4
    ph_setr = np.expand_dims(ph_setr,axis=time_dim)
    # make individual-channel magnitudes and phases 
    magn_idch = np.absolute(varr.data)
    ph_idch = np.arctan2(np.imag(varr.data),np.real(varr.data)) 
    #TODO this is for testing
    # save individual channels
    """
    varr_ph = copy.copy(varr)
    varr_setr = copy.copy(varr)
    varr_setr.data = ph_setr
    varr_ph.data = ph_idch
    ph_test_out = workdir+'/ph_test'
    ph_setr_out = workdir+'/ph_setr'
    vj.core.write_nifti(varr_ph,ph_test_out,\
                rcvr_to_timedim=True,combine_rcvrs=False)
    vj.core.write_nifti(varr_setr,ph_setr_out,\
                rcvr_to_timedim=True,combine_rcvrs=False)
    """
    # creating summed output
    csum = np.sum(magn_idch*np.exp(1j*(ph_idch-ph_setr)),axis=rcvr_dim)
    # final composer phase
    comp_ph = np.arctan2(np.imag(csum),np.real(csum))
    # magnitude weighted output
    comp_magn_w = np.sqrt(np.abs(np.sum(\
                magn_idch**2*np.exp(1j*(ph_idch-ph_setr)),axis=rcvr_dim)))
    comp_cmplx = comp_magn_w*np.exp(1j*comp_ph)
    # expand to rcvr_dim
    comp_cmplx = np.expand_dims(comp_cmplx,axis=-1)
    #update data
    varr.data = comp_cmplx
    if keepfiles==True:
        vj.core.write_nifti(varr,composer_out,\
            save_complex=True,rcvr_to_timedim=True,combine_rcvrs=False)
    elif keepfiles == False:
        shutil.rmtree(workdir)
    varr.description = 'phasematched'
    return varr




import glob
from setuptools import setup, find_packages

with open('README.md','r') as fh:
    long_description = fh.read()

script_list = glob.glob('vnmrjpy/scripts/*')

setup(
    name='vnmrjpy',
    version='0.1.dev3',
    author='David Hlatky',
    author_email='hlatkydavid@gmail.com',
    url='https://github.com/hlatkydavid/vnmrjpy',
    description='Handle VnmrJ MRI data and recostruction with Python',
    long_description=long_description,
    python_requires='>=3.5',
    #packages=setuptools.find_packages(exclude=\
    #['dataset','test','.test.*','*.test.*',\
    #'vnmrsys','vnmrjpy.bin','vnmrjpy.recon','vnmrjpy.io']),
    packages=find_packages(exclude=['test']),
    include_package_data=True,
    scripts=script_list,
    install_requires=[
        'lmfit',
        'numpy',
        'scipy',
        'matplotlib',
        'nibabel',
        ],
)



import vnmrjpy as vj
import numpy as np
import copy

"""
Contains functions to transform vnmrjpy.varray into other coordinate systems.
These functions are not intented to be called on their own, but from the
appropriate method of a varray object. More documentation is present at
vnmrjpy/core/varray.py
"""
def _to_scanner(varr):
    """Transform data to scanner coordinate space by properly swapping axes

    Standard vnmrj orientation - meaning rotations are 0,0,0 - is axial, with
    x,y,z axes (as global gradient axes) corresponding to phase, readout, slice.
    vnmrjpy defaults to handling numpy arrays of:
                (receivers, phase, readout, slice/phase2, time*echo)
    but arrays of
                        (receivers, x,y,z, time*echo)
    is also desirable in some cases (for example registration in FSL flirt)

    Euler angles of rotations are psi, phi, theta,

    Also corrects reversed X gradient and sliceorder

    Args:
        data (3,4, or 5D np ndarray) -- input data to transform
        procpar (path/to/file) -- Varian procpar file of data

    Return:
        swapped_data (np.ndarray)
    """
    return varr

def _to_anatomical(varr):

    if varr.space == 'anatomical':
        return varr
    if _check90deg(varr.pd) == False:
        raise(Exception('Only multiples of 90 deg allowed here.'))    
    # TODO this is the old method
    #new_sdims, flipaxes = _anatomical_sdims(varr.pd['orient'])

    # better method is via rotation matrix
    swapaxes, flipaxes = _anatomical_swaps(varr.pd)
    
    # setting data: 90deg rotation is done by swapping and flipping
    oldaxes = [i for i in range(varr.data.ndim)]
    newaxes = copy.copy(oldaxes)
    newaxes[0:3] = swapaxes
    #varr.data = np.moveaxis(varr.data, newaxes, oldaxes) 
    varr.data = np.moveaxis(varr.data, oldaxes, newaxes) 
    
    # ------flipping as part of rotation----------------------
    varr.data = _flip_axes(varr.data,flipaxes)

    # setting sdims
    #varr = _move_sdims(varr,new_sdims) 
    sdims_partial = varr.sdims[:3]
    sdims_kept = varr.sdims[3:]
    new_sdims_partial = [i[1] for i in sorted(zip(swapaxes, sdims_partial))]
    
    new_sdims = new_sdims_partial+sdims_kept

    varr.space = 'anatomical'
    varr.set_dims()
    varr.sdims = new_sdims
    # -----additional custom flip data on axes------------------

    #this is 
    #varr.flip_axis('z')
    #varr.flip_axis('phase')
    #varr.flip_axis('slice')
    
    # 3D should be flipped. Dunno why, but seems to work
    if varr.pd['apptype'] in ['im3D','im3Dfse']:
        varr.flip_axis('slice')

    # update header
    varr.set_nifti_header()

    return varr

def _to_global(varr):
    raise(Exception('Not implemented'))
    return varr

def _to_local(varr):
    """Transform to [phase, read, slice, etc..]

    Sets appropriate varray attributes including nifti_header
    If it is first called from to_kspace() automatically, and also from other
    'read' methods
    """
    #TODO autocall from other read functions
    # if space is None, then the call is from to_kspace()
    if varr.space == None:
    
        varr.space = 'local'
        varr.sdims = ['phase','read','slice','time','rcvr']
        varr.set_dims()
        if vj.config['swap_x_grad'] == True:
            varr.flip_axis('x')
        varr.set_nifti_header()

    elif varr.space == 'anatomical':
        raise(Exception('not implemented yet'))
    elif varr.space == 'local':
        # do nothing, already in local space
        return varr 

    return varr

def _check90deg(pd):
    """Return True if the Euler angles are 0,90,180, etc"""
    psi, phi, theta = int(pd['psi']), int(pd['phi']), int(pd['theta'])
    if psi % 90 == 0 and phi % 90 == 0 and theta % 90 == 0:
        return True
    else:
        return False

def _flip_axes(data,axes):
    """Return flipped data on axes"""
    for ax in axes:
        data = np.flip(data, axis=ax)
    return data

def _anatomical_swaps(pd):
    """Return swap and flip arrays for data transform to anatomical

    use_hardcoded: no-brain implementation for 90deg rots
    """
    use_hardcoded = True
    # hardcoded for 90degs
    if use_hardcoded:
        if _check90deg(pd) != True:
            raise(Exception('Not implemented'))

        ori = pd['orient']
        if ori == 'trans':
            #swap = [0,1,2]
            swap = [0,2,1]
            flip = [1,2]
        elif ori == 'trans90':
            #swap = [1,0,2]
            swap = [2,0,1]
            flip = [0]
        elif ori == 'sag':
            #swap = [1,2,0]
            swap = [2,1,0]
            flip = [1,2]
        elif ori == 'sag90':
            #swap = [2,1,0]
            swap = [1,2,0]
            flip = [0]
        elif ori == 'cor':
            #swap = [0,2,1]
            swap = [0,1,2]
            flip = [1]
        elif ori == 'cor90':
            swap = [1,0,2]
            flip = []

        return swap, flip
    # with rot matrix
    else:
        rot_matrix = vj.core.niftitools._qform_rot_matrix(pd)
        inv = np.linalg.inv(rot_matrix).astype(int)
        swap = inv.dot(np.array([1,2,3], dtype=int))
        flipaxes = []
        for num, i in enumerate(swap):
            if i < 0:
                flipaxes.append(num)
        swapaxes = (np.abs(swap) - 1).astype(int)

    return swapaxes, flipaxes
    

def _move_sdims(varr,sdims_new):
    """Move varray data aexes according to new sdims"""

    # move
    # set new sdims
    varr.sdims = sdims_new
    return varr


def _set_dims(varr):
    """Set varray dims"""
    if varr.space == 'local' or varr.space == None:
        if varr.dims != None:
            return varr
        if _check90deg(varr.pd) != True: 
            varr.dims = 'oblique'
            return varr
        swapax, flipax = _anatomical_swaps(varr.pd)
        swap = swapax
        anat_dims = ['x','y','z']
        dims = [i[1] for i in sorted(zip(swap,anat_dims))]
        varr.dims = dims
        return varr

    elif varr.space == 'anatomical':
        # TODO set dims only if 90 deg rot case
        if _check90deg(varr.pd): 
            varr.dims = ['x','y','z']
        else:
            varr.dims = 'oblique'
        return varr

def _flip_axis(varr,axis):
    """Flip data on axis 'x','y','z' or 'phase','read','slice'"""
    if axis in varr.dims:
        ax = varr.dims.index(axis)
        varr.data = np.flip(varr.data,axis=ax)
    elif axis in varr.sdims:
        ax = varr.sdims.index(axis)
        varr.data = np.flip(varr.data,axis=ax)
    else:
        raise(Exception('Axis not specified correctly'))
    
    return varr



import unittest
import vnmrjpy  as vj
import imageio
import numpy as np
import sys

# working, but convergence is hard...

class Test_SVT(unittest.TestCase):
    
    def test_svt_func(self):
        
        # read boat image
        image = '/boat.png'
        thresh = 0.8
        img = imageio.imread(vj.pics+image)
        mask = np.random.rand(img.shape[0],img.shape[1])
        mask[mask >= thresh] = 1
        mask[mask < thresh] = 0
        img_masked = np.multiply(img, mask)
        filled = vj.aloha.lowranksolvers.svt(img_masked,known_data=img_masked,\
                                            tau=50000,realtimeplot=True)
        #svt = vj.aloha.SingularValueThresholding(img_masked,realtimeplot=True,\
        #                                        tau=50000,max_iter=1000,delta=1)
        #svt.solve()
        
        



import numpy as np
import vnmrjpy as vj

class KspaceCompleter():
    """Finalize kspace. Use ALOHA for completion, or pass if not using cs.

    K-space completion is done in the ALOHA framework. Low rank matrix
    completion parameters are initialized from vnmrjpy.config and procpar
    """
    def __init__(self,kspace,procpar):
        """Init ALOHA 'reconpar' and other params"""
        ppdict = vj.io.ProcparReader(procpar).read()

        if ppdict['pslabel'] == 'ge3d_elliptical':
            self.is_complete = True
        elif 'skipint' in ppdict.keys():
            self.is_complete = False
        else:
            self.is_complete = True

        self.kspace = kspace
        self.ppdict = ppdict
        self.procpar = procpar

    def make(self):
        """Run ALOHA or pass if already complete"""

        if self.is_complete:
            return self.kspace
        else:
            # TODO
            raise(Exception('Not implemented yet'))
            # TODO if reconpar is in procpar
            #rp = self.ppdict['reconpar']
            rp = None

            aloha = vj.aloha.Aloha(self.kspace,self.procpar,reconpar=rp)
            return aloha.recon()
    
    def to_global(self):
        pass            



import vnmrjpy as vj
import unittest
import numpy as np
import nibabel as nib
import glob
import copy

# test on small number of slices only for 'speed'
SLC = 'all'
SLCNUM = 1

class Test_Aloha(unittest.TestCase):

    """
    def test_angio(self):
    
        kspace_orig, kspace_cs, affine, procpar, savedir = \
                                        self._load_data(data='angio')

        aloha = vj.aloha.Aloha(kspace_cs,procpar)
        kspace_filled = aloha.recon() 
        self._save_test_results(procpar,affine,savedir,\
                                kspace_orig,kspace_cs,kspace_filled)
    """
    def test_mems(self):
    
        kspace_orig, kspace_cs, affine, procpar, savedir = \
                                        self._load_data(data='mems')

        aloha = vj.aloha.Aloha(kspace_cs,procpar)
        kspace_filled = aloha.recon() 
        self._save_test_results(procpar,affine,savedir,\
                                kspace_orig,kspace_cs,kspace_filled)
    def test_gems(self):
    
        kspace_orig, kspace_cs, affine, procpar, savedir = \
                                        self._load_data(data='gems')

        nib.viewers.OrthoSlicer3D(np.absolute(kspace_cs)[0,...]).show()
        aloha = vj.aloha.Aloha(kspace_cs,procpar)
        kspace_filled = aloha.recon() 
        self._save_test_results(procpar,affine,savedir,\
                                kspace_orig,kspace_cs,kspace_filled)
    """
    def test_gems(self):
    
        kspace_orig, kspace_cs, affine, procpar, savedir = \
                                        self._load_data(data='ge3d')

        nib.viewers.OrthoSlicer3D(np.absolute(kspace_cs)[0,...]).show()
        aloha = vj.aloha.Aloha(kspace_cs,procpar)
        kspace_filled = aloha.recon() 
        self._save_test_results(procpar,affine,savedir,\
                                kspace_orig,kspace_cs,kspace_filled)
    """
    #---------------------------------UTILITY----------------------------------
    def _save_test_results(self,procpar,\
                                affine,\
                                savedir,\
                                kspace_orig,\
                                kspace_cs,\
                                kspace_filled):

        k_name = ['kspace_orig','kspace_cs','kspace_filled']
        img_name = ['img_orig','img_cs','img_filled']
        dirs = ['orig','zerofilled','filled']
        for num, item in enumerate([kspace_orig,kspace_cs,kspace_filled]):
            recon= vj.recon.ImageSpaceMaker(item, procpar)
            img = recon.make()
            saverecon = vj.io.SaveRecon(procpar,kspace=item,imagespace=img)
            outdir = savedir+'/'+dirs[num]
            saverecon.save(outdir,savetype='full',filetype='nifti')


    def _load_data(self,data=None):

        if data == 'mems':
            seqdir = '/mems_s_2018111301_axial_0_0_0_01.cs'
        elif data == 'angio':
            seqdir = '/ge3d_angio_HD_s_2018072604_HD_01.cs'
        elif data == 'gems':
            seqdir = '/gems_s_2018111301_axial90_0_90_0_01.cs'
        elif data == 'ge3d':
            seqdir = '/ge3d_s_2018111301_axial_0_0_0_01.cs'
        elif data == None:
            raise(Exception('Testdata not specified'))

        testdir = vj.cs+seqdir
        procpar = testdir + '/procpar'
        resultdir = vj.config['testresults_dir']+'/aloha'+seqdir[:-3]+'.nifti'
        savedir = resultdir

        imag = []
        real = []
        imag_orig = []
        real_orig = []
    
        mask_img = nib.load(glob.glob(testdir+'/*mask.nii.gz')[0])
        affine = mask_img.affine
        mask = mask_img.get_fdata()

        for item in sorted(glob.glob(testdir+'/*kspace*_imag*')):
            data = nib.load(item).get_fdata()
            imag_orig.append(data)
            imag.append(np.multiply(data,mask))
        for item in sorted(glob.glob(testdir+'/*kspace*_real*')):
            data = nib.load(item).get_fdata()
            real_orig.append(data)
            real.append(np.multiply(data,mask))
        #for item in [imag,real,imag_orig,real_orig]:
        #    item = np.asarray(item)
        imag = np.asarray(imag)
        real = np.asarray(real)
        imag_orig = np.asarray(imag_orig)
        real_orig = np.asarray(real_orig)
        kspace_cs = np.vectorize(complex)(real,imag)
        kspace_orig = np.vectorize(complex)(real_orig,imag_orig)
   
        if SLC == 'all':
            return (kspace_orig,\
                    kspace_cs,\
                    affine,\
                    procpar,\
                    savedir)
        else:
            return (kspace_orig[...,SLC:SLC+SLCNUM,:],\
                    kspace_cs[...,SLC:SLC+SLCNUM,:],\
                    affine,\
                    procpar,\
                    savedir)



import copy
import numpy as np
import vnmrjpy as vj
import sys
np.set_printoptions(formatter={'float': lambda x: "{0:0.3f}".format(x)})
"""
Collection of solvers for low rank matrix completion
"""

class SingularValueThresholding():
    """
    Matrix completion by singular value soft thresholding
    """
    def __init__(self,A,tau=None, delta=None, epsilon=1e-4, max_iter=1000,\
                realtimeplot=True):

        self.A = A
        self.Y = np.zeros_like(A)
        self.max_iter = max_iter
        self.epsilon = epsilon
        mask = copy.deepcopy(A)
        mask[mask != 0] = 1
        self.mask = mask 
        if tau == None:
            self.tau = 5*np.sum(A.shape)/2
        else:
            self.tau = tau
        if delta == None:
            self.delta = 1.2*np.prod(A.shape)/np.sum(self.mask)
        else:
            self.delta = delta

        if realtimeplot == True:
            self.rtplot = vj.util.RealTimeImshow(np.absolute(A))
            self.realtimeplot = realtimeplot

    def solve(self):
        """Main iteration, returns the completed matrix"""
        for _ in range(self.max_iter):

            U, S, V = np.linalg.svd(self.Y, full_matrices=False)
            S = np.maximum(S-self.tau, 0)
            X = np.linalg.multi_dot([U, np.diag(S), V])
            self.Y = self.Y + self.delta*self.mask*(self.A-X)
        
            rel_recon_error = np.linalg.norm(self.mask*(X-self.A)) / \
                                np.linalg.norm(self.mask*self.A)

            if _ % 100 == 0:
                sys.stdout.flush()
                print(rel_recon_error)
                pass
            if rel_recon_error < self.epsilon:
                break
            if self.realtimeplot == True:
                self.rtplot.update_data(np.absolute(X))
        return X



import vnmrjpy as vj
import numpy as np
import copy

"""
Basic functions on varrays.
Includes:
    noise_mask
    average
    
"""

def mask(varr, threshold=None):
    """Create mask on image: set voxel data to 0 where image is below threshold.

    Noise threshold is assumed if set to None

    """ 
    return varr

def average(*varr_list):

    pass



import vnmrjpy as vj
from vnmrjpy.core.utils import vprint
import numpy as np
import nibabel as nib
import glob
from shutil import copyfile
import warnings
import os

def write_fdf(varray, out):
    """Write fdf files from varray data into .img direcotory specified in out"""
    pass

def write_nifti(varr,out, save_procpar=True,\
                        save_complex=False,\
                        rcvr_to_timedim=False,\
                        complex_pairs='magn-phase',\
                        combine_rcvrs='ssos',\
                        cut_to_3d=False):
    """Write Nifti1 files from vnmrjpy.varray.
  
    You can specify the data to be saved 
 
    Args:
        varray       -- vnmrjpy.varray object containing the data
        out          -- output path
        save_procpar -- (boolean) saves procpar json in the same directory
        save_complex -- save complex data in real-imag pairs along a dimension
        only4dim     -- compress data into 4 dimensions , if possible
        
    """
    # sanity check for input options

    if combine_rcvrs !=None and save_complex == True:
        #TODO
        pass

    # check type of data
    if len(varr.data.shape) < 3:
        raise(Exception('Data dimension error'))
    if not varr.is_kspace_complete:
        raise(Exception('K-space is not completed'))
    #------------------ making the Nifti affine and header-----------------

    # main write
    if '.nii' in out:
        out_name = str(out)
    elif '.nii.gz' in out:
        out_name = str(out[:-3])
    else:
        out_name = str(out)+'.nii'
  
    data = varr.data 
    
    # OPTION : save_complex----------------------------------------------------

    # to save complex data, real and imaginary parts are added
    # to receiver channel
    if save_complex:
        rch = 4

        if complex_pairs =='real-imag':
            data = np.concatenate([np.real(data), np.imag(data)],axis=rch)
        elif complex_pairs == 'magn-phase':
            magn = np.absolute(data)
            phase = np.arctan2(np.imag(data),np.real(data))
            data = np.concatenate([magn,phase],axis=rch)
        else:
            raise(Exception('Wrong complex_pairs'))

    # OPTION : rcvr_to_timedim-------------------------------------------------

    if rcvr_to_timedim:
        rch = 4
        timedim = data.shape[rch]*data.shape[3]
        newshape = (data.shape[0],data.shape[1],data.shape[2],timedim)
        data = np.reshape(data,newshape, order='c')

    # OPTION : combine_rcvrs---------------------------------------------------

    if combine_rcvrs == 'ssos':

        data = vj.core.recon.ssos(data,axis=4)
        # cut dimensions more than 3. This is mainly for use with certain FSL calls
        if cut_to_3d:
            if len(data.shape) == 4:
                data = data[...,0]
            elif len(data.shape) == 5:
                data = data[...,0,0]
            else:
                raise(Exception('Not implemented'))
    # set affine as none to use the one in header
    img = nib.Nifti1Image(data, None, varr.nifti_header)
    # create directories if not exists
    basedir = out_name.rsplit('/',1)[0]
    if not os.path.exists(basedir):
        os.makedirs(basedir)
    nib.save(img,out_name)
    os.system('gzip -f '+str(out_name))
    vprint('write_nifti : '+str(out_name)+' saved ... ')

    if save_procpar:
        new_pp = out_name[:-3]+'procpar'
        try:
            copyfile(varr.procpar, new_pp)
        except:
            pdname = out_name[:-3]+'json'
            vj.core.utils.savepd(varr.pd,pdname)
            # TODO consider saving only in json
            #warnings.warn('Could not copy procpar. Saved dictionary as json.')




import vnmrjpy as vj
import numpy as np
import unittest
import glob
import os
import matplotlib.pyplot as plt


class Test_kmake_fsems(unittest.TestCase):

    def test_kmake_from_fid(self):

        fsems_dir = sorted(glob.glob(vj.fids+'/fsems_s_2018111301_axial_0*'))[0]
        fid = fsems_dir+'/fid'
        procpar = fsems_dir+'/procpar'
        fiddata, fidhdr = vj.io.FidReader(fsems_dir).read()
        
        kmaker = vj.recon.KspaceMaker(fiddata, fidhdr, procpar, verbose=True) 
        kspace = kmaker.make()
        
        out_dir = vj.config['testresults_dir']+'/kmake/fsems'
        outfile_name_k = 'fsems_test_kspace'
        outfile_name_i = 'fsems_test_imgspace'
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        outpath_k = out_dir+'/'+outfile_name_k
        outpath_i = out_dir+'/'+outfile_name_i

        imaker = vj.recon.ImageSpaceMaker(kspace, procpar)
        imgspace = imaker.make()
        writer = vj.io.NiftiWriter(np.absolute(kspace[1,...]),procpar)
        writer.write(outpath_k)
        writer = vj.io.NiftiWriter(np.absolute(imgspace[1,...]),procpar)
        writer.write(outpath_i)




import vnmrjpy as vj
import numpy as np
import time
import os
import matplotlib.pyplot as plt
from nipype.interfaces import fsl

class Composer():
    """Match the phases of different receiver channels

    Ref.: Robinson et.al: Combining Phase Images from  Array Coils Using a
          Short Echo Time Reference Scan (COMPOSER)

    """
    def __init__(self, imgspace, imgspace_ref,procpar, procpar_ref,\
                    rcvrdim='default',workdir=None):
        """
        Args:
            imgspace (np.ndarray) -- complex image data in numpy array
            imgspace_ref (np.ndarray) -- complex reference image data
            procpar

        """
        # moving receiverdim to last, and keeping there from now on
        if rcvrdim == 'default':
            # if input is from default layout switch rcvrdim to dim 4
            self.rcvrdim = 4
            self.imgspace = np.moveaxis(imgspace,[1,2,3,4,0],[0,1,2,3,4])
            self.imgspace_ref = np.moveaxis(imgspace_ref,[1,2,3,4,0],[0,1,2,3,4])
        else:
            raise(Exception('not implemented'))
        self.procpar = procpar
        self.procpar_ref = procpar_ref

        if workdir == None:
            timestr = time.strftime("%Y%m%d-%H%M%S")
            #self.workdir = os.path.expanduser('~')+'/tmp_vnmrjpy/cmp'+timestr
            self.workdir = vj.config['fsl_workdir']+'/composer_'+timestr
        else:
            self.workdir = workdir
        # temporary workdir for testing
        self.workdir = '/home/david/dev/vnmrjpy/test/results/composertest'


    def flirt_registration(self, keepfiles=False):
        """

        """
        def _ssos(data):
            """Squared sum of squares from multichannel complex data"""
            ssos = np.sqrt(np.mean(np.absolute(data),axis=self.rcvrdim))
            return ssos

        # make workdir if does not exist
        if not os.path.exists(self.workdir):
            os.makedirs(self.workdir)

        # make 6d nifti data for flirt affine transform
        img_6d = vj.util.make_6d(self.imgspace)
        # making niftis
        img_out = self.workdir+'/composer_img_main'
        ref_out = self.workdir+'/composer_ref_main'
        flirt_out = self.workdir+'/composer_flirtout'
        full = self.workdir + '/composer_6d_img'

        writer = vj.io.NiftiWriter(_ssos(self.imgspace), self.procpar,\
                                        input_space='local',\
                                        output_space='scanner')
        writer.write(img_out)
        writer = vj.io.NiftiWriter(_ssos(self.imgspace_ref), self.procpar_ref,\
                                        input_space='local',\
                                        output_space='scanner')
        writer.write(ref_out)
        writer = vj.io.NiftiWriter(img_6d, self.procpar_ref,\
                                        input_space='local',\
                                        output_space='scanner')
        writer.write(full)

        # registering by combined magnitude FSL FLIRT
        flirt = fsl.FLIRT()
        flirt.inputs.in_file = img_out+'.nii.gz'
        flirt.inputs.reference = ref_out+'.nii.gz'
        flirt.inputs.out_file = flirt_out+'.nii.gz'
        flirt.inputs.out_matrix_file = 'composer_invol2refvolmat'
        print(flirt.cmdline)
        res = flirt.run()
        # applying the resulting transform to full data
        flirt = fsl.FLIRT()
        flirt.inputs.in_file = img_out+'.nii.gz'
        flirt.inputs.reference = ref_out+'.nii.gz'
        flirt.inputs.out_file = flirt_out+'.nii.gz'
        flirt.inputs.out_matrix_file = 'composer_invol2refvolmat'
        print(flirt.cmdline)
        vj.io.NiftiReader(flirt_out+'.nii.gz')
        # loading result nifti

        rdr = vj.io.NiftiReader(flirt_out+'.nii.gz')

    def match_from_nifti():
        pass







import numpy as np
import nibabel as nib
import vnmrjpy as vj


class NiftiReader():
    """Reads nifti with nibabel, and does some checks


    """
    def __init__(self,nifti,input_space='rat_anatomical'):
        
        self.nifti = nifti
        self.input_space = input_space

    def read():

        n = nib.load(self.nifti)
        header = n.header
        procpar = header['aux_file']
        data = n.get_fdata()
        if procpar == '':
            print('Warning! procpar could not be found based on aux_file')

        return data, header, procpar



import numpy as np
import vnmrjpy as vj
from scipy.ndimage import median_filter
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt
import copy
import sys

#TODO getting R2 error

class T2Fitter():

    def __init__(self,img_data, echo_times=[],\
                                procpar=None,\
                                skip_first_echo=False,\
                                mask=None,\
                                automask=False,\
                                fitmethod='scipy',\
                                fitfunc='3param_exp'):

        """Fit exponential to multiecho data 

        Args:
            img_data (np.ndarray) -- 4D imput data in image space
            echo_times (tuple) -- echo times corresponding to dim4 in img_data
            skip_first_echo (boolean) -- skips first datapoint if needed
            mask (np.ndarray) -- binary mask where fitting should be done
            automask (boolean) -- tries to detect noise and make proper mask
            fitmethod (str) -- 
            fitfunc (str) -- 
        """
        # if no echo time is specified get from procpar
        def _check_echo_times(tdim, echo_list):
            if len(echo_list) == tdim and tdim != 0:
                return True
            else:
                return False

        def _get_noise_mean(img_data):
            """Find mean noise by making a histogram"""

            # def constants
            bins = 100

            imgmax = np.max(np.array(img_data,dtype=np.float64))
            hist, bin_edges = np.histogram(img_data, bins=bins, range=(0,imgmax))
            # probably most of image is noise
            noise_peak = np.amax(hist)
            noise_mean = np.where(hist==noise_peak)[0][0]*bins
            return noise_mean

        def _automask(img_data):
            """Make mask by thresholding for mean noise and median filtering"""

            filter_size = 7
            
            noise_thresh = _get_noise_mean(img_data)*3
            mask = np.ones_like(img_data[...,0])
            mask[img_data[...,0] < noise_thresh] = 0
            mask = median_filter(mask, size=filter_size)
            # tile mask to same size as img_data
            mask = np.repeat(mask[...,np.newaxis],img_data.shape[-1],axis=-1)
            return mask

        # check data dimensions (should be 4)
        if len(img_data.shape) != 4:
            raise(Exception('Input data should have 4 dimensions')) 
        # sanity check for fitfunc
        if fitfunc in ['3param_exp','2param_exp']:
            pass
        else:
            raise(Exception('Cannot recognize fitfunc'))

        if _check_echo_times(img_data.shape[-1],echo_times) == False \
                                                    and procpar != None:
            ppdict = vj.io.ProcparReader(procpar).read()
            ne = int(ppdict['ne'])
            te = float(ppdict['te'])
            echo_times = [te * i for i in range(1,ne+1)]
        else:
            raise(Exception('Please specify procpar or echo times'))

        if skip_first_echo:
            self.data = img_data[:,:,:,1:]  # time dim is 4
            self.echo_times = echo_times[1:]
        else:
            self.data = img_data
            self.echo_times = echo_times
       
        if mask == None and automask == True:
            mask = _automask(self.data)
        else:
            mask = np.ones_like(self.data)
 
        self.mask = mask
        self.fitmethod = fitmethod  # some string option for later
        self.fitfunc = fitfunc
        self.data = np.array(self.mask * self.data,dtype=np.float64)
        self.echo_times = np.array(self.echo_times,dtype=np.float64)
        self.noise_mean = _get_noise_mean(img_data)

    def fit(self): 
        """Actual fitting method

        Return:
            (*fitted parameters)        
        """
        def fit_2param_exp(data,echo_times, mask):

            def _2param_exp(x,T2,A):
                y = A*np.exp(-x*1000/(T2*1000))
                return y

            t2map = np.zeros_like(data[...,0])
            m0map = np.zeros_like(data[...,0])
            et = self.echo_times
            nm = self.noise_mean
            y_err = np.asarray([nm/5 for i in range(len(echo_times))])
            print('Fitting 2 parameter exponential')
            for z in range(data.shape[2]):
                for y in range(data.shape[1]):
                    for x in range(data.shape[0]):
                        if mask[x,y,z,0] == 0:
                            continue
                        else:
                            try:
                                data_1d = data[x,y,z,:]
                                y0 = data_1d[0]
                                (t2,m0), cov  = curve_fit(_2param_exp,\
                                                et,\
                                                data_1d,\
                                                p0=(0.03,y0*1.5),\
                                                check_finite=False,\
                                                bounds=([0,0],[0.2,y0*3]),\
                                                method='trf')
                                                #sigma=y_err,\
                                                #absolute_sigma=True)
                                m0map[x,y,z] = m0
                                t2map[x,y,z] = t2
                                x_fit = np.linspace(0,et[-1],100)
                            except Exception:
                                t2map[x,y,z] = guess[0]
                                m0map[x,y,z] = guess[1]
                            except KeyboardInterrupt:
                                sys.exit(0)
            return (t2map, m0map)

        def fit_3param_exp(data,echo_times, mask):
            """Standard 3 parameter wxponential fitting"""

            def _3param_exp(x,T2,A,B):
                y = A*np.exp(-x*1000/(T2*1000)) + B
                return y

            t2map = np.zeros_like(data[...,0])
            m0map = np.zeros_like(data[...,0])
            const = np.zeros_like(data[...,0])
            
            et = self.echo_times
            nm = self.noise_mean
            # init and bounds for estimated params B, T2,
            #(A depends on intensity)
            y_err = np.asarray([nm for i in range(len(echo_times))])
            print('Fitting 3 parameter exponential')
            for z in range(data.shape[2]):
                for y in range(data.shape[1]):
                    for x in range(data.shape[0]):
                        if mask[x,y,z,0] == 0:
                            continue
                        else:
                            data_1d = data[x,y,z,:]
                            y0 = data_1d[0]
                            guess = (0.03,y0*1.5,nm)
                            bounds = ([0,y0/2,nm/2],[0.5,y0*2,nm*3])
                            try:
                                (t2,m0,c), cov  = curve_fit(_3param_exp,\
                                            et,\
                                            data_1d,\
                                            p0=guess,\
                                            check_finite=False,\
                                            bounds=bounds,\
                                            method='trf',\
                                            sigma=y_err,\
                                            absolute_sigma=True)
                                t2map[x,y,z] = t2
                                m0map[x,y,z] = m0
                                const[x,y,z] = c
                            except Exception:
                                t2map[x,y,z] = guess[0]
                                m0map[x,y,z] = guess[1]
                                const[x,y,z] = guess[2]
                            except KeyboardInterrupt:
                                sys.exit(0)
            return (t2map, m0map, const)

        #---------------------------Switch cases ------------------------------

        if self.fitfunc == '3param_exp':        
            
            return  fit_3param_exp(self.data,self.echo_times,self.mask)

        elif self.fitfunc == '2param_exp':        
            
            return fit_2param_exp(self.data,self.echo_times,self.mask)

















import vnmrjpy as vj
import unittest
import matplotlib.pyplot as plt
import numpy as np


class Test_AlohaTest_fiberload(unittest.TestCase):

    def test_load_test_cs_fiber(self):

        alohatest = vj.util.AlohaTest('offcenter',1)
        fiber_orig, fiber_cs = alohatest.load_test_cs_fiber('gems',recontype='k')
        
        imgmax = np.max(np.absolute(fiber_orig))

        plt.subplot(1,2,1)
        plt.imshow(np.absolute(fiber_orig[0,...]),cmap='gray', vmin=0,vmax=10)
        plt.title('original abs')
        plt.subplot(1,2,2)
        plt.imshow(np.absolute(fiber_cs[0,...]),cmap='gray', vmin=0,vmax=10)
        plt.title('cs abs')
        plt.show()

         



import vnmrjpy as vj
import numpy as np
import unittest
import glob
import matplotlib.pyplot as plt

class Test_Wasabi(unittest.TestCase):

    def test_wasabi(self):

        seq = glob.glob(vj.config['dataset_dir']+\
                        '/parameterfit/wasabi/gems*test05*.fid')[0]
        seq_norm = glob.glob(vj.config['dataset_dir']+\
                        '/parameterfit/wasabi/gems*norm02*fid')[0]

        varr = vj.core.read_fid(seq).to_kspace()
        varr.to_imagespace().to_anatomical()
        varr_norm = vj.core.read_fid(seq_norm).to_kspace()
        varr_norm.to_imagespace().to_anatomical()
        # reduce data along slice dim to make test faster
        mask = vj.core.recon.ssos(varr_norm.data)
        # naive mask for testing
        mask[mask < 5] = 0
        mask[mask != 0] = 1
        (res, err) = vj.func.WASABI(varr, varr_norm, mask=mask) 
        out_b0 = vj.config['testresults_dir']+'/wasabi/b0map'
        out_b1 = vj.config['testresults_dir']+'/wasabi/b1map'
        out_c = vj.config['testresults_dir']+'/wasabi/cmap'
        out_d = vj.config['testresults_dir']+'/wasabi/dmap'
        # errors
        out_b0_err = vj.config['testresults_dir']+'/wasabi/b0map_err'
        out_b1_err = vj.config['testresults_dir']+'/wasabi/b1map_err'
        out_c_err = vj.config['testresults_dir']+'/wasabi/cmap_err'
        out_d_err = vj.config['testresults_dir']+'/wasabi/dmap_err'

        vj.core.write_nifti(res[0],out_b0)
        vj.core.write_nifti(res[1],out_b1)
        vj.core.write_nifti(res[2],out_c)
        vj.core.write_nifti(res[3],out_d)

        vj.core.write_nifti(err[0],out_b0_err)
        vj.core.write_nifti(err[1],out_b1_err)
        vj.core.write_nifti(err[2],out_c_err)
        vj.core.write_nifti(err[3],out_d_err)



import os
import unittest
import vnmrjpy as vj
import glob

class Test_fid2nii(unittest.TestCase):

    def test_fid2nii(self):

        outdir = vj.config['testresults_dir']+'/bin'
        fid = glob.glob(vj.fids+'/gems*axial_0_0_0*')[0]
        out = outdir+'/'+fid.rsplit('/')[-1][:-4]
        os.system('fid2nii '+fid+' '+out)



import unittest
import vnmrjpy as vj
import numpy as np
import matplotlib.pyplot as plt


class Test_Admm(unittest.TestCase):

    def test_on_mems_data(self):

        alohatest = vj.util.AlohaTest('offcenter',1)
        fiber_orig, fiber_cs = alohatest.load_test_cs_fiber('mems',recontype='k-t')
        rp = {'recontpye':'k-t','filter_size':(11,5),'vcboost':False,\
                'stage':3,'rcvrs':4,'fiber_shape':fiber_cs.shape}
        hankel_cs = vj.aloha.construct_hankel_2d(fiber_cs, rp)
        hankel_orig = vj.aloha.construct_hankel_2d(fiber_orig,rp)
        print('starting lmafit')
        X,Y,out = vj.aloha.lowranksolvers.lmafit(hankel_cs, \
                                            known_data=hankel_cs,\
                                            realtimeplot=False)
        stage=0
        print('starting admm')
        hankel_finished = vj.aloha.lowranksolvers.admm(X,Y.conj().T,\
                                                    fiber_cs,stage,rp,\
                                                    realtimeplot=True,\
                                                    max_iter=200,\
                                                    mu=1000,\
                                                    verbose=True)
        
    """
    def test_run_admm(self):
        # this is just to check if it runs...
        shape = (4,192,192)
        rp = {'recontype':'kx-ky','filter_size':(17,17),'vcboost':False,\
                'stage':3,'rcvrs':4,'fiber_shape':shape}
        # TODO load actual data
        stage = 0
        d1 = np.random.rand(*shape)
        d2 = np.random.rand(*shape)
        fiber_orig  = np.vectorize(complex)(d1,d2)
        mask = np.ones(fiber_orig.shape)
        mask[fiber_orig < 0.7] = 0
        fiber_known = np.multiply(fiber_orig,mask)
        hankel_known = vj.aloha.construct_hankel(fiber_known,rp)
        X,Y,out = vj.aloha.lowranksolvers.lmafit(hankel_known, \
                                            known_data=hankel_known,\
                                            realtimeplot=False)
        hankel_finished = vj.aloha.lowranksolvers.admm(X,Y.conj().T,\
                                                    fiber_orig,stage,rp,\
                                                    realtimeplot=True)
    """
    

    



import vnmrjpy as vj
import numpy as np
from math import sin, cos


def make_cubic_affine(data):
    """Make Nifti affine assuming imaging volume is a cube

    Args:
        data (np.ndarray) -- 3D or 4D input data to make affine for
    Return:
        affine (np.ndarray) -- 4x4 or 3x3? diagonal matrix

    Useful for easy viewing, but don!t use it for serios work.
    """
    x = 1/data.shape[0]
    y = 1/data.shape[1]
    z = 1/data.shape[2]
    dim_arr = np.array([x,y,z,1])
    affine = np.eye(4)*dim_arr
    return affine

def make_rat_anatomical_affine(procpar):
    """Make rat_anatomical affine for Nifti header.

    Almost the same as make_scanner_affine

    Args:
        data (np.ndarra) -- input data, 4D or 5D
        procpar 
    Return:
        affine (np.ndarray) -- 4x4 diagonal matrix
    """

    p = vj.io.ProcparReader(procpar).read()
    #arr =  dimensions of (phase, read, slice/phase2)
    arr = vj.util.get_local_pixdim(p)
    # params to transform to scanner from local
    swaparr, flipaxis, sliceaxis = get_swap_array(p['orient'])
    sortedarr = sorted(zip(swaparr,arr))
    arr = [sortedarr[i][1] for i in range(3)]
    # params to trnasform from scanner to rat_anat
    
    swaparr2rat = [0,2,1] 
    sortedarr2 = sorted(zip(swaparr2rat,arr))
    arr = [sortedarr2[i][1] for i in range(3)]
    arr = arr + [1]
    #TODO make translations
    affine = np.array(np.eye(4,dtype=float)*arr,dtype=float)
    return affine

def make_scanner_affine(procpar):
    """Make appropriate scanner affine for Nifti header.

    This is needed for viewing and saving nifti in scanner space
    Prabably safer to do it from scratch than transforming...
    Translations are accounted for. Gaps are added to thinckness
    
    Args:
        data (np.ndarra) -- input data, 4D or 5D
        procpar 
    Return:
        affine (np.ndarray) -- 4x4 diagonal matrix
    """
    p = vj.io.ProcparReader(procpar).read()
    orient = p['orient']
    arr = get_local_pixdim(p)        
    swaparr, flipaxis, sliceaxis = get_swap_array(p['orient'])
    sortedarr = sorted(zip(swaparr,arr))
    arr = [sortedarr[i][1] for i in range(3)]+[1.0]
    #TODO make translations
    affine = np.array(np.eye(4,dtype=float)*arr,dtype=float)
    return affine
    
def scanner_to_rat_anatomical(data):
    """Transform from scanner to rat anatomical

    """
    datadims = len(data.shape)
    olddims = [0,1,2]  # scanner [x,y,z]
    newdims = [0,2,1]
    if datadims == 4:
        olddims = olddims+[3]
        newdims = newdims+[3]
    if datadims == 5:
        olddims = [0,1,2,3,4]
        newdims = [0,1,3,2,4]
        # change axis
    if datadims == 6:  # [x,y,z,time,rcvrs,etc]
        olddims = [0,1,2,3,4,5]
        newdims = [0,2,1,3,4,5]
    data = np.moveaxis(data,olddims, newdims)
    return data

def to_rat_anatomical_space(indata, procpar, flip_gradients=True):
    """Transform data to rat_anatomical coordinates, similar to scanner space.

    Args:
        indata
        procpar
        flip_gradients
    Return:
        data -- data with swapped axes

    """
    data = to_scanner_space(indata,procpar,flip_gradients)
    return scanner_to_rat_anatomical(data)

def to_scanner_space(indata, procpar, flip_gradients=True):
    """Transform data to scanner coordinate space by properly swapping axes

    Standard vnmrj orientation - meaning rotations are 0,0,0 - is axial, with
    x,y,z axes (as global gradient axes) corresponding to phase, readout, slice.
    vnmrjpy defaults to handling numpy arrays of:
                (receivers, phase, readout, slice/phase2, time*echo)
    but arrays of
                        (receivers, x,y,z, time*echo)
    is also desirable in some cases (for example registration in FSL flirt)

    Euler angles of rotations are psi, phi, theta,

    Also corrects reversed X gradient and sliceorder

    Args:
        data (3,4, or 5D np ndarray) -- input data to transform
        procpar (path/to/file) -- Varian procpar file of data

    Return:
        swapped_data (np.ndarray)
    """

    def _orthotransform(data,p):
        """Simple swaping of axes in case of orthogonal scanner-image axes

        This was done to avoid thinking. Would benefit from rewriting.
        """

        orient = p['orient']
        sliceorder = int(p['sliceorder'])
        dims = len(data.shape)
        newarr, flipaxis, sliceaxis = get_swap_array(orient)  
        #flipping slices before axis transform
        if '2D' in p['apptype'] or '2d' in p['apptype']:
            # sliceorder can be 0,1,2,3 odd means interleaved,
            # > 2 means reversed order
            if int(p['sliceorder']) < 2:
                data = flip_sliceaxis(data)         

        # flipping gradients before axis transform
        if flip_gradients==True:
        
            data = flip_peaxis(data)
            data = flip_roaxis(data)
            if '3D' in p['apptype'] or '3d' in p['apptype']:
                data = flip_pe2axis(data)
        # moving around axes according to orient
        if dims == 3:
            data = np.moveaxis(data, [0,1,2], newarr)
        elif dims == 4:
            newarr = newarr+[3]
            data = np.moveaxis(data, [0,1,2,3], newarr)
            #data = np.moveaxis(data, newarr,[0,1,2,3])
        elif dims == 5:
            if flipaxis != None:
                flipaxis = np.array(flipaxis)+1
            newarr = [i+1 for i in newarr]
            newarr = [0]+newarr+[4]
            data = np.moveaxis(data, [0,1,2,3,4], newarr)
        elif dims == 6:
            if flipaxis != None:
                flipaxis = flipaxis
            newarr = newarr+[3,4,5]
            print(newarr)
            data = np.moveaxis(data, [0,1,2,3,4,5], newarr)
        #flipping axes according to orient
        if flipaxis != None:
            for i in flipaxis:
                data = np.flip(data, axis=i) 

        
        # correct in case X gradient is inverted
        #data = corr_x_flip(data)
        # correct reversed slice order
        return data
    #--------------------------------------------------------------------------
    #TODO this is for later, for more sophisticated uses perhaps
    def _calc_matrix(p):
        """calculate rotation matrix"""

        psi = int(p['psi'])/360*(2*np.pi)
        phi = int(p['phi'])/360*(2*np.pi)
        theta = int(p['theta'])/360*(2*np.pi)
        # rotation matrix with the euler angles
        R = [[cos(phi)*cos(psi)-sin(phi)*cos(theta)*sin(psi),\
                    -sin(phi)*cos(psi)-sin(psi)*cos(theta)*cos(phi),\
                    sin(theta)*sin(psi)],\
            [cos(phi)*sin(psi)+sin(phi)*cos(theta)*cos(psi),\
                    -sin(phi)*sin(psi)+cos(psi)*cos(theta)*cos(phi),\
                    -sin(theta)*cos(psi)],\
            [sin(theta)*sin(phi), sin(theta)*cos(phi), cos(theta)]]
    #-------------------------------main---------------------------------------
    if len(indata.shape) in [3,4,5,6]:
        pass
    else:
        raise(Exception('Only 3d, 4d or 5d data allowed'))
    p = vj.io.ProcparReader(procpar).read()

    if check_90deg(procpar) == True:
        
        return _orthotransform(indata, p)
    
    else:
        raise(Exception('Only supported with rotations of multiples of 90deg'))

    #TODO transform affine as well
    

def check_90deg(procpar):
    """Make sure all rotations are multiples of 90deg

    Euler angles of rotations are psi, phi, theta,

    Args:
        procpar
    Return:
        True or False

    """
    p = vj.io.ProcparReader(procpar).read()
    psi, phi, theta = int(p['psi']), int(p['phi']), int(p['theta'])
    if psi % 90 == 0 and phi % 90 == 0 and theta % 90 == 0:
        return True
    else:
        return False

def flip_roaxis(data,axis='default'):

    if axis=='default':
        if len(data.shape) < 5:
            ax = vj.config['ro_dim']-1
        elif len(data.shape) == 5:
            ax = vj.config['ro_dim']
        elif len(data.shape) > 5:
            ax = vj.config['ro_dim']-1
    else:
        ax = axis
    return np.flip(data,axis=ax)

def flip_peaxis(data,axis='default'):

    if axis=='default':
        if len(data.shape) < 5:
            ax = vj.config['pe_dim']-1
        elif len(data.shape) == 5:
            ax = vj.config['pe_dim']
        elif len(data.shape) > 5:
            ax = vj.config['pe_dim']-1
    else:
        ax = axis
    return np.flip(data,axis=ax)

def flip_sliceaxis(data,axis='default'):
        
    if axis=='default':
        if len(data.shape) < 5:
            ax = vj.config['slc_dim']-1
        elif len(data.shape) == 5:
            ax = vj.config['slc_dim']
        elif len(data.shape) > 5:
            ax = vj.config['slc_dim']-1
    else:
        ax = axis
    return np.flip(data,axis=ax)
        
def flip_pe2axis(data,axis='default'):

    if axis=='default':
        if len(data.shape) < 5:
            ax = vj.config['pe2_dim']-1
        elif len(data.shape) == 5:
            ax = vj.config['pe2_dim']
        elif len(data.shape) > 5:
            ax = vj.config['pe2_dim']-1
    else:
        ax = axis
    return np.flip(data,axis=ax)

def corr_sliceaxis(data,sliceaxis,sliceorder):            
    """Flip slice after transform to scanner space

    """
    # this is reversed now: actually flip if 'rev slices' is unchecked

    # flip slice, so increasing order is up->down, left->right,s->n
    if sliceorder < 2:  # this means 'rev order' is unchecked
        if len(data.shape) < 5:
            data = np.flip(data,axis=sliceaxis)
        elif len(data.shape) == 5:
            data = np.flip(data,axis=(sliceaxis+1))
        elif len(data.shape) > 5:
            data = np.flip(data,axis=sliceaxis)
    return data

def corr_x_flip(data):
    """Check and correct X gradient polarity flip"""
    if vj.config['swap_x_grad']:
        if len(data.shape) < 5:
            data = np.flip(data,axis=0)
        elif len(data.shape) == 5:
            data = np.flip(data,axis=1)
        elif len(data.shape) > 5:
            data = np.flip(data,axis=0)
    return data

def get_swap_array(orient):
    """Gives how to reorder and flip the axes to fit scanner space

    [phase, readout, slice] -> [x,y,z]
    
    Used with np.moveaxes

    Args:
        orient -- procpar parameter 'orient'
    Returns:
        arr -- sequence of new axes
        flipaxis -- axis to flip
        sliceaxis -- axis of slice encoding
    """
    if orient == 'trans':
        arr = [0,1,2]
        flipaxis = None
        sliceaxis = 2
    elif orient == 'trans90':
        arr = [1,0,2]
        flipaxis = [1]
        sliceaxis = 2
    elif orient == 'sag':
        arr = [1,2,0]
        flipaxis = None
        sliceaxis = 0
    elif orient == 'sag90': 
        arr = [2,1,0]
        flipaxis = [2]
        sliceaxis = 0
    elif orient == 'cor':
        arr = [0,2,1]
        flipaxis = [1]
        sliceaxis = 1
    elif orient == 'cor90': 
        arr = [1,2,0]
        arr = [2,0,1]
        flipaxis = [1,2]
        sliceaxis = 1
    else:
        raise(Exception('Other orientations not implemented'))

    return arr, flipaxis, sliceaxis

def get_local_pixdim(p):
    """Calculate pixel dimension in (pe,ro,slc)

    Args:
        p -- procpar ditionary
    Returns:
        arr -- 1x3 array of dimensions
    """
    mul = 10
    if 'epip' not in p['pslabel']:
        phase = int(p['np'])//2
        read = int(p['nv'])
        # these are in cm!
        lro = float(p['lro'])*mul
        lpe = float(p['lpe'])*mul
        if '2D' in p['apptype']:
            slices = int(p['ns'])
            thk = float(p['thk'])
            gap = float(p['gap'])
            arr = [lpe/phase,lro/read,(thk+gap)]
        elif '3D' in p['apptype']:
            phase2 = int(p['nv2'])
            lpe2 = float(p['lpe2'])*mul
            arr = [lpe/phase,lro/read,lpe2/phase2]
        else:
            raise(Exception)
    else:
        raise(Exception)

    return arr



import unittest
import vnmrjpy as vj
import numpy as np

class Test_Pyramidal(unittest.TestCase):

    def test_pyramidal_kt(self):

        rp = {'rcvrs':4,'fiber_shape':(4,128,21),'recontype':'k-t',\
                'stages':3,'solver':'lmafit','virtualcoilboost':False,\
                'filter_size':(11,7)} 
        a = np.random.rand(4,128,21)
        b = np.random.rand(4,128,21)
        kspace_fiber = np.vectorize(complex)(a,b)
        weights = vj.aloha.make_kspace_weights(rp)
        kspace_fin = vj.aloha.pyramidal_kt(kspace_fiber,weights,rp)

        self.assertEqual(kspace_fin.shape,kspace_fiber.shape)

    def test_pyramidal_kxky(self):

        rp = {'rcvrs':4,'fiber_shape':(4,128,128),'recontype':'kx-ky',\
                'stages':3,'solver':'lmafit','virtualcoilboost':False,\
                'filter_size':(11,11)} 
        a = np.random.rand(4,128,128)
        b = np.random.rand(4,128,128)
        kspace_fiber = np.vectorize(complex)(a,b)
        weights = vj.aloha.make_kspace_weights(rp)
        kspace_fin = vj.aloha.pyramidal_kxky(kspace_fiber,weights,rp)

        self.assertEqual(kspace_fin.shape,kspace_fiber.shape)
